/****************************************************************************

    t6963c.c
    --------

	controlling a T6963C graphics display

    Author:		Jens Renner
    Date:    	2004-01-02
	Updated:	2004-01-05
    Software: 	AVR-GCC

	Hardware:
	- 240x64 pixel
	- Data lines on port C, control line on port A (pins 0..4) - see t6963c.h
	- 6x8 hard wired
	- testet with AT90S8535 (@ 4 MHz) and ATmega163 (@ 8 MHz)

	History:
	2004-01-05: finished 6x8 mode in lcd_bitmap(), works for full-screen images
	2004-01-04: fixed some bugs in display mode functions.
				text & graphics mode are working
	2004-01-03: display reset works
	
	To-do:
	- 8X8 mode
	- implement lcd_txtcursor()/lcd_gfxcursor() functions
	- write some basic graphics functions (line, circle, rectangle ...)

****************************************************************************/

#define __AVR_AT90S8535__

#include <avr/io.h>		// access to controller I/O
#include <inttypes.h>	// integer types
#include <progmem.h>	// access to program memory
#include "t6963c.h"		// definitions and prototypes


static volatile uint8_t lcd_mode = LCD_CMD_SETDISPLAYMODE; // save current display mode (text/graphics/none/both)

// place 240x64 image (1-bit depth) in program memory
uint8_t __attribute__ ((progmem)) lcd_bitmap[1920] = {
	0xFF, 0xFF, 0xFF, 0xF4, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x55, 0x55, 0x49, 0x24, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF,
	0xAB, 0x5A, 0xAA, 0xAA, 0xAA, 0xAA, 0xAA, 0xAA, 0xAA, 0xAA, 0xAA, 0xAA, 0xAA, 0xAA, 0xAA, 0xAA,
	0xAA, 0xAA, 0xAA, 0xAA, 0x12, 0x00, 0x24, 0x91, 0x55, 0xB6, 0xDD, 0xBE, 0xBF, 0xFE, 0xED, 0xA4,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0A, 0xC0, 0x00,
	0x01, 0x12, 0xAA, 0xFF, 0xB7, 0x6F, 0xFF, 0xFF, 0xF7, 0xF7, 0xFF, 0xFF, 0xB6, 0xA8, 0x92, 0x4A,
	0x52, 0xAA, 0xAA, 0xAA, 0xAA, 0xAA, 0xAA, 0xAA, 0xAA, 0xAA, 0xAA, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF, 0xDB, 0xFF, 0xFF, 0x7D, 0xBB, 0x7F, 0xDF, 0xDF, 0xFD, 0x54, 0x92, 0x24, 0x90, 0x84, 0x00,
	0x22, 0x44, 0x44, 0x95, 0x55, 0x5B, 0x77, 0xFF, 0xFF, 0xFF, 0xFF, 0xDF, 0xFE, 0xFB, 0xBD, 0xFF,
	0xFD, 0xFD, 0xEF, 0xFF, 0xFD, 0xFD, 0x20, 0x02, 0x42, 0x41, 0x11, 0x05, 0x2A, 0xDD, 0xDD, 0xBB,
	0xFF, 0xFF, 0xFF, 0xFF, 0xFE, 0xDB, 0xBF, 0x7F, 0xFE, 0xFB, 0xBF, 0xDF, 0xF7, 0xEE, 0xDF, 0xDF,
	0xFE, 0xDB, 0xAF, 0x6F, 0x4A, 0xA9, 0x3A, 0xBE, 0xEE, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFE, 0xF7,
	0x7F, 0xFF, 0xDF, 0xFF, 0xF7, 0xEF, 0xFF, 0xFF, 0xF7, 0x7D, 0xBE, 0xBB, 0x6A, 0xB5, 0x25, 0x6A,
	0x51, 0x29, 0xBF, 0xFF, 0xEF, 0xFF, 0xFF, 0xFE, 0xF7, 0x77, 0x77, 0xFF, 0x6F, 0xDF, 0xED, 0xDB,
	0x7B, 0xFD, 0xDD, 0xFB, 0xEB, 0x55, 0x59, 0xA4, 0x81, 0x4A, 0xAD, 0xAA, 0xDB, 0x55, 0xBD, 0xAA,
	0xFB, 0x7F, 0xFF, 0xED, 0xFF, 0xF7, 0xDF, 0xFF, 0xFF, 0x6D, 0xFB, 0xFD, 0xBF, 0x7D, 0xEE, 0x96,
	0x6A, 0x9F, 0xED, 0x55, 0x56, 0xB7, 0x7E, 0xB5, 0x76, 0xB5, 0x55, 0x55, 0x42, 0x55, 0xFF, 0xED,
	0xFE, 0xFF, 0xB6, 0xDF, 0xFD, 0xB6, 0xDB, 0xF7, 0x55, 0x25, 0x51, 0x45, 0x2A, 0xD5, 0xAE, 0xF2,
	0x80, 0x5D, 0x55, 0x55, 0x49, 0x4A, 0x89, 0x55, 0xB6, 0xEE, 0xFF, 0xFF, 0xDB, 0xFF, 0x57, 0xB6,
	0xED, 0xB4, 0x44, 0x89, 0x24, 0x95, 0x5A, 0xDB, 0x6E, 0xB5, 0xAA, 0xB6, 0xB2, 0xB9, 0x7F, 0x77,
	0xAB, 0x6D, 0xB7, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xDB, 0xB6, 0x92, 0xA8, 0x89, 0x12, 0x4B,
	0x5B, 0x75, 0xBB, 0x6A, 0xAB, 0x6D, 0x55, 0x54, 0xAA, 0x95, 0x5B, 0x78, 0xEB, 0xA8, 0xFF, 0xFF,
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x6A, 0xA9, 0x55, 0x55, 0x5A, 0xD7, 0x76, 0xED, 0xB5, 0x6A, 0x95,
	0x4D, 0x55, 0x55, 0x55, 0xB6, 0xEF, 0xFF, 0xFF, 0xFF, 0xF9, 0xBD, 0xD8, 0x3F, 0xFF, 0xFE, 0xAA,
	0xFF, 0xB7, 0x77, 0xFF, 0xFF, 0xFE, 0xAA, 0xAB, 0x59, 0x55, 0x2A, 0x4A, 0x95, 0x6D, 0xB6, 0xFF,
	0xFF, 0xFF, 0xFF, 0xFF, 0xDF, 0xFF, 0xFF, 0xBE, 0xD2, 0xA8, 0x0A, 0xAA, 0xAB, 0xFF, 0xFF, 0xF9,
	0xDD, 0x5F, 0xD7, 0xC2, 0x55, 0x54, 0xA5, 0x55, 0xDB, 0xFF, 0x7F, 0xFF, 0xFF, 0xDF, 0xFF, 0xFF,
	0xFF, 0x76, 0xF5, 0x6A, 0xDB, 0xFD, 0xE8, 0xA4, 0x03, 0xF7, 0x75, 0x49, 0xFF, 0xE0, 0xA0, 0x1F,
	0xEF, 0xFF, 0xB6, 0xFF, 0xFF, 0xFF, 0xFF, 0xFB, 0xFF, 0xFF, 0xFF, 0xFA, 0xAA, 0x92, 0x01, 0x9B,
	0x5F, 0xDF, 0xB6, 0xBB, 0x28, 0x00, 0x00, 0x1A, 0xDF, 0xBF, 0xFF, 0xD3, 0xDD, 0x5F, 0xFD, 0xBF,
	0xFF, 0xEF, 0xBF, 0xFF, 0x7E, 0xAE, 0xD5, 0x55, 0x5B, 0xBF, 0xFF, 0x7D, 0xFE, 0xED, 0x6A, 0x69,
	0x5A, 0xBD, 0x00, 0x00, 0x00, 0x1F, 0xF6, 0xFB, 0xFF, 0xFF, 0xFF, 0xF5, 0xAB, 0xF5, 0xFF, 0xBD,
	0xEA, 0xAA, 0xD5, 0xFB, 0xFF, 0xFB, 0xED, 0x64, 0x92, 0xA6, 0xAB, 0x56, 0xDD, 0xDC, 0x0F, 0xF6,
	0x00, 0x80, 0x00, 0xA3, 0x7F, 0xEF, 0xFF, 0xFF, 0xFF, 0xFE, 0xFF, 0x7F, 0xDB, 0xF6, 0xBF, 0xDD,
	0xB6, 0xAA, 0x24, 0x4A, 0x35, 0xBB, 0xEF, 0xFF, 0xFF, 0xFF, 0xF7, 0xF6, 0x72, 0xDE, 0xB9, 0xB0,
	0x9D, 0x60, 0xDB, 0x7D, 0xFF, 0xFF, 0xDA, 0xFF, 0xFF, 0xE0, 0x6C, 0x95, 0x50, 0x52, 0xAA, 0xAF,
	0xFB, 0xFF, 0xFF, 0xEF, 0x7D, 0xBB, 0x6D, 0xBB, 0x7E, 0xAC, 0x25, 0x76, 0x08, 0xF6, 0xED, 0xE0,
	0x24, 0x82, 0xFF, 0xFF, 0x7A, 0xFB, 0x40, 0x2E, 0x93, 0x6F, 0xEF, 0xFF, 0xFF, 0xFB, 0x6F, 0xAD,
	0xDF, 0x7F, 0xF7, 0xEF, 0xFF, 0xFF, 0xDB, 0xE0, 0x01, 0x16, 0x80, 0xF9, 0xB5, 0xB0, 0x0A, 0xBA,
	0xFF, 0xFF, 0xEF, 0xDA, 0x3B, 0x69, 0xFF, 0xFD, 0x7F, 0xAE, 0xDB, 0xBF, 0xFE, 0xFF, 0xFB, 0xFD,
	0xDF, 0x7D, 0xB5, 0x55, 0x54, 0x98, 0x1E, 0x8A, 0x00, 0xAD, 0xDD, 0xB0, 0x02, 0xA5, 0xFF, 0xFF,
	0xFF, 0xFA, 0xDB, 0x7F, 0xDD, 0xB7, 0xDB, 0xFB, 0xFE, 0xF6, 0xDB, 0xF6, 0xDE, 0xD6, 0xB4, 0x92,
	0x49, 0x08, 0xAB, 0x50, 0x7E, 0xCF, 0x80, 0xAA, 0x68, 0xD0, 0x00, 0xDA, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF, 0xFB, 0xFF, 0xFF, 0xFE, 0xDF, 0x6F, 0xDB, 0x6D, 0x09, 0x22, 0x55, 0x4A, 0xD5, 0xAD, 0xF7,
	0x55, 0x7E, 0x6E, 0xCA, 0x80, 0x00, 0x00, 0x00, 0x00, 0x7F, 0xFF, 0xFF, 0xFF, 0xBF, 0xFF, 0xEE,
	0xB6, 0xAA, 0xAA, 0xA4, 0x80, 0x04, 0x92, 0xED, 0x5D, 0xAF, 0x6B, 0x56, 0xAA, 0x08, 0xAA, 0xFD,
	0xBF, 0xC5, 0x00, 0x01, 0x40, 0x13, 0xE4, 0x1F, 0xFF, 0xFF, 0x5B, 0xFF, 0xFA, 0xBB, 0x51, 0x10,
	0x12, 0x55, 0x7B, 0x75, 0x6D, 0x2A, 0xA5, 0x55, 0x29, 0x55, 0x55, 0xFF, 0xFF, 0x7F, 0xFE, 0xD4,
	0x80, 0x10, 0xAD, 0x57, 0xFF, 0x87, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xEA, 0xAD, 0xD7, 0xD5, 0xAA,
	0x8A, 0x95, 0x12, 0x92, 0xBA, 0xDF, 0xFF, 0xFE, 0xFF, 0xF7, 0x68, 0xFF, 0xB7, 0xFF, 0xD0, 0x90,
	0xB2, 0xAF, 0xFF, 0x81, 0xFF, 0xFF, 0xFF, 0x7A, 0xFF, 0xFF, 0x54, 0x28, 0x22, 0x15, 0x6F, 0xFF,
	0xFF, 0xFF, 0xFF, 0xFD, 0xDF, 0x70, 0x43, 0x7F, 0x8F, 0x80, 0xFF, 0xE5, 0x7F, 0x54, 0x54, 0xAF,
	0xFF, 0x80, 0x7F, 0xFF, 0xFF, 0xEF, 0xBF, 0xFB, 0x4B, 0xAF, 0xDF, 0xFF, 0xFD, 0xFF, 0xFB, 0xFF,
	0xD7, 0xFF, 0xFA, 0x80, 0x05, 0xD4, 0xFB, 0x65, 0xDE, 0xFD, 0x3F, 0xFB, 0x00, 0x03, 0xCE, 0x00,
	0x3F, 0xFF, 0xFB, 0xE5, 0xFF, 0xFF, 0xFE, 0xFB, 0xFF, 0xFF, 0xFF, 0xF7, 0xEE, 0x80, 0x03, 0xF7,
	0xFF, 0x00, 0x00, 0x0A, 0x4B, 0xF7, 0x77, 0xFF, 0xF7, 0xEF, 0xC0, 0x01, 0xC5, 0x15, 0xF7, 0xFF,
	0xFF, 0x9F, 0xFF, 0xEA, 0xFB, 0xFF, 0x7A, 0xD5, 0xBF, 0x7E, 0xF8, 0x00, 0x0F, 0xBF, 0xFF, 0xD7,
	0xEA, 0xFF, 0xFE, 0xFF, 0xFF, 0x6F, 0xDF, 0x7E, 0xFD, 0x43, 0xCF, 0xFF, 0xFF, 0xFF, 0xFF, 0x7F,
	0xFF, 0xEA, 0xDF, 0xAD, 0xD7, 0xBF, 0xFB, 0xD5, 0x00, 0x08, 0x3F, 0xFF, 0xFF, 0xFF, 0xFF, 0xED,
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFD, 0xFF, 0xFF, 0xEE, 0xFF, 0xF7, 0x7F, 0xFF, 0xF5,
	0xF6, 0xF6, 0xBA, 0xA7, 0xFE, 0xC0, 0x00, 0x00, 0x3F, 0xFF, 0xFF, 0xFF, 0xBF, 0xBF, 0xFF, 0xFF,
	0xFB, 0xF5, 0x5A, 0xDF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFE, 0xF7, 0xF7, 0xFA, 0x08, 0x09,
	0x4A, 0xDF, 0xF7, 0x00, 0x00, 0x00, 0xFF, 0xFB, 0xFF, 0xFD, 0xFF, 0xFF, 0xFF, 0xFF, 0xD5, 0x00,
	0x00, 0x01, 0x7F, 0xFF, 0xFF, 0xFF, 0x6F, 0xFF, 0xFE, 0xFF, 0xFF, 0xFB, 0xF7, 0xFF, 0xFF, 0xFF,
	0xD8, 0x00, 0x00, 0x00, 0xFF, 0x6F, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xE8, 0x00, 0x00, 0x00,
	0x07, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFD, 0xDF, 0xFF, 0xFF, 0xDF, 0xF7, 0x77, 0x7F, 0xC0, 0x00,
	0x00, 0x01, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xAF, 0xB6, 0xE7, 0x00, 0x04, 0x00, 0xBF,
	0xFF, 0xFF, 0xBF, 0xFF, 0xFE, 0xFF, 0xFD, 0xFF, 0xFD, 0xBF, 0xDF, 0xFE, 0x00, 0x00, 0x00, 0x2F,
	0xFF, 0xDF, 0x7B, 0xFF, 0xFF, 0xFF, 0xFE, 0xB7, 0xCE, 0x62, 0x00, 0x04, 0x00, 0x5F, 0xFF, 0xFF,
	0xFF, 0xFF, 0xFD, 0xFF, 0xFF, 0xFF, 0xB7, 0xED, 0x7A, 0xB4, 0x00, 0x00, 0x02, 0xBF, 0xFF, 0xBD,
	0xFF, 0xFF, 0xFF, 0xFF, 0xFD, 0x5D, 0xCE, 0x72, 0x00, 0x0E, 0x00, 0x3F, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF, 0xFF, 0xFE, 0xFD, 0xDD, 0x55, 0x4A, 0xB0, 0x00, 0x01, 0x5F, 0xFF, 0xFF, 0x7B, 0xFF, 0x7F,
	0xBF, 0xFF, 0xFE, 0xD7, 0xA6, 0x70, 0xF3, 0x6F, 0x80, 0x5F, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF, 0xFF, 0x52, 0xAA, 0xAA, 0x00, 0x00, 0x2D, 0x7F, 0xFF, 0xFF, 0xBF, 0xFF, 0xFF, 0xF7, 0xFF,
	0xFA, 0xBF, 0xC7, 0x75, 0x9A, 0x3F, 0x00, 0x1F, 0xFF, 0xFF, 0xFF, 0xFF, 0xFD, 0xBF, 0xFE, 0xFF,
	0x4A, 0xAA, 0xA8, 0x00, 0x0A, 0xD7, 0xFF, 0xFF, 0xFE, 0xF7, 0xFF, 0xFF, 0xAF, 0xFF, 0xFD, 0x6B,
	0xC7, 0x7D, 0x9B, 0xFD, 0x00, 0x3F, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFD, 0xFD, 0x24, 0x12,
	0xA8, 0x02, 0xAA, 0xBA, 0xFF, 0xFF, 0xFF, 0x7F, 0xFF, 0xDF, 0x8F, 0xFF, 0xFA, 0xBF, 0xA3, 0x9D,
	0xFB, 0xEF, 0xBF, 0xDF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFD, 0xFE, 0xFF, 0xDB, 0xFF, 0xF0, 0x35,
	0x6D, 0xEF, 0xFF, 0xFF, 0xFF, 0x7F, 0xFF, 0xFD, 0xAF, 0xFF, 0xFF, 0x57, 0xC3, 0x99, 0x87, 0xFC,
	0x26, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFA, 0xEC, 0xFF, 0xFF, 0xA5, 0xDF, 0xBF, 0xFF,
	0xFF, 0xFF, 0xFE, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFD, 0xFB, 0xA1, 0x99, 0x9E, 0xAC, 0x9B, 0xFF,
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xBB, 0xFA, 0xD9, 0xFF, 0xDB, 0x7F, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF, 0xEF, 0xFF, 0xDE, 0xEF, 0xFF, 0xFF, 0xAF, 0xC1, 0x88, 0xFB, 0xCE, 0x0B, 0xFF, 0xFF, 0xFF,
	0xFF, 0xFF, 0xFF, 0xFF, 0xF1, 0xFC, 0xFF, 0x7E, 0xEF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFE, 0xFF,
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xE1, 0x08, 0xEC, 0x00, 0x26, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF, 0xEF, 0xF5, 0xB9, 0xEF, 0xFB, 0x7F, 0xFB, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xD7, 0x5D,
	0xFE, 0xFF, 0xFF, 0xFF, 0xB0, 0x03, 0x20, 0x00, 0x17, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	0xCB, 0xBA, 0xBF, 0xFD, 0xB6, 0xBF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x7F, 0xDF, 0xDE, 0xFF, 0xFF,
	0xFF, 0xFF, 0xE0, 0x00, 0x00, 0x00, 0x5F, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xF5, 0xB7, 0x74,
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFE, 0xFF, 0x5D, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFC, 0x00, 0x01, 0x2F, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFE, 0x9F, 0x75, 0xFF, 0xFF,
	0xFF, 0xDB, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xF7, 0xDE, 0xDF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFD, 0xF4, 0xBF, 0xFF, 0xFF, 0x7F,
	0x7F, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFE, 0xBB, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xF5, 0xE2, 0xFF, 0xFF, 0xFF, 0xDF, 0xF7, 0xFF,
	0xFF, 0xFF, 0xFF, 0xFB, 0xFD, 0x7D, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xD7, 0xD5, 0xFF, 0xFF, 0xFF, 0x7F, 0xF2, 0xFF, 0xFF, 0xFF,
	0xFF, 0xFF, 0xD6, 0xF7, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF, 0xFF, 0xFF, 0xFF, 0x7F, 0xAA, 0x1F, 0xFA, 0xFF, 0xFF, 0xF2, 0xED, 0xFF, 0xFF, 0xFF, 0xFE,
	0xF9, 0xF7, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF, 0xFF, 0xFF, 0x55, 0x3D, 0x5F, 0x7F, 0xFF, 0x71, 0x3B, 0x5F, 0xFF, 0xFF, 0xFF, 0xAF, 0xEF,
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFE, 0xEA, 0x5F, 0xFF, 0xFF, 0xBF, 0xF0, 0x55, 0x47, 0xFF, 0xFF, 0xFF, 0xFF, 0xDF, 0xFF, 0xFF,
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xB5,
	0x7F, 0xFB, 0xFF, 0xFF, 0xF5, 0x58, 0xDF, 0xFF, 0xFF, 0xFF, 0xFF, 0xBF, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xB2, 0x4A, 0x15, 0x4A,
	0xAA, 0xEB, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x6D, 0x54, 0x94, 0x80, 0x49, 0x20, 0xD5, 0x55, 0x55, 0x2A,
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF, 0xB5, 0x55, 0x21, 0x52, 0xAA, 0xA2, 0x5A, 0x94, 0x8A, 0x24, 0xAA, 0xAA, 0xD5, 0x7F, 0xFF,
	0xFD, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xDA, 0xAA, 0x40, 0xAA,
	0xAA, 0x95, 0x29, 0x21, 0x15, 0x22, 0x42, 0x21, 0xAA, 0xAA, 0xAB, 0x6F, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFD, 0xDA, 0xAA, 0xAA, 0x48, 0x25, 0x55, 0x2D, 0x49, 0x48, 0xAA,
	0x84, 0x94, 0xA2, 0x49, 0x28, 0x94, 0x55, 0x55, 0xBF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF, 0xFF, 0xB5, 0x26, 0x55, 0x55, 0x52, 0xA5, 0xAA, 0x48, 0xA2, 0x24, 0xA5, 0x24, 0x55, 0x4A,
	0x14, 0x84, 0x85, 0x02, 0xAF, 0xFF, 0xFF, 0xFF, 0xBF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xAA,
	0xD5, 0x55, 0x55, 0x52, 0x49, 0x2A, 0x11, 0x25, 0x15, 0x52, 0x12, 0x91, 0x20, 0x21, 0x42, 0x28,
	0x50, 0xA9, 0x55, 0x5F, 0xFF, 0xFF, 0xFF, 0xFF, 0xEE, 0xDB, 0x7F, 0xFF, 0x6A, 0xAA, 0x55, 0x52,
	0x92, 0x49, 0x24, 0x91, 0x4A, 0xA9, 0x48, 0x09, 0x48, 0x44, 0x95, 0x48, 0x28, 0x93, 0x2A, 0xAA };



/***** LCD_READSTATUS *******************************************************
	reads the display status
****************************************************************************/
int lcd_readstatus (void) {
	uint8_t status;
	LCD_DATA_DDR = 0; // DDR: input
	LCD_DATA_OUT = 0xff; // enable internal pull-ups
	sbi(LCD_CTRL_OUT, LCD_CTRL_PIN_CD); // this is a command
	sbi(LCD_CTRL_OUT, LCD_CTRL_PIN_WR); // no write
	cbi(LCD_CTRL_OUT, LCD_CTRL_PIN_RD); // read
	cbi(LCD_CTRL_OUT, LCD_CTRL_PIN_CE); // chip enable
	asm volatile( // wait
    "nop"	"\n\t"
    "nop"	"\n\t"
    "nop"	"\n\t"
    "nop"
    );
	status = LCD_DATA_IN; // data in
	sbi(LCD_CTRL_OUT, LCD_CTRL_PIN_CE); // chip disable
	return (status);
}


/***** LCD_WAIT *************************************************************
	waits until STA0 and STA1 are set
****************************************************************************/
void lcd_wait (void) {
	uint8_t status;
	status = lcd_readstatus(); // read status byte
	while( (status&0x03) != 0x03 ){ // both bits set?
		status = lcd_readstatus(); // check again
	}

}


/***** LCD_WRITE ************************************************************
	writes a byte into the display's data buffer
****************************************************************************/
void lcd_write(uint8_t data) {
	lcd_wait(); // check status
	LCD_DATA_OUT = data; // data out
	LCD_DATA_DDR = 0xff; // DDR: output
	cbi(LCD_CTRL_OUT, LCD_CTRL_PIN_CD); // this is data
	cbi(LCD_CTRL_OUT, LCD_CTRL_PIN_WR); // write
	sbi(LCD_CTRL_OUT, LCD_CTRL_PIN_RD); // no read
	cbi(LCD_CTRL_OUT, LCD_CTRL_PIN_CE); // chip enable
	asm volatile( // data setup time
    "nop"	"\n\t"
    "nop"	"\n\t"
    "nop"	"\n\t"
    "nop"
    );
	sbi(LCD_CTRL_OUT, LCD_CTRL_PIN_CE); // chip disable
	asm volatile( // data hold time
    "nop"
    );
}



/***** LCD_COMMAND **********************************************************
	sends a command
****************************************************************************/
void lcd_command(uint8_t cmd) {
	lcd_wait(); // check status
	LCD_DATA_OUT = cmd; // data out
	LCD_DATA_DDR = 0xff; // DDR: output
	sbi(LCD_CTRL_OUT, LCD_CTRL_PIN_CD); // this is a command
	cbi(LCD_CTRL_OUT, LCD_CTRL_PIN_WR); // write
	sbi(LCD_CTRL_OUT, LCD_CTRL_PIN_RD); // no read
	cbi(LCD_CTRL_OUT, LCD_CTRL_PIN_CE); // chip enable
	asm volatile( // data setup time
    "nop"	"\n\t"
    "nop"	"\n\t"
    "nop"	"\n\t"
    "nop"
    );
	sbi(LCD_CTRL_OUT, LCD_CTRL_PIN_CE); // chip disable
	asm volatile( // data hold time
    "nop"
    );
}


/***** LCD_TXTDATA **********************************************************
	sends an ascii character
****************************************************************************/
void lcd_txtdata(uint8_t chr) {
	chr -= 0x20; // -32 for compatibility reasons
	lcd_write(chr);
	lcd_command(LCD_CMD_SENDDATAINC);
}


/***** LCD_GFXDATA **********************************************************
	sends 6/8 pixel
****************************************************************************/
void lcd_gfxdata(uint8_t pixel) {
	lcd_write(pixel);
	lcd_command(LCD_CMD_SENDDATAINC);
}


/***** LCD_SETTXTMODE *******************************************************
	enables text mode
****************************************************************************/
void lcd_settxtmode(void) {
	lcd_mode = lcd_mode | LCD_CMD_SETDISPLAYMODE_TXT; // set bit
	lcd_command(lcd_mode);
}


/***** LCD_CLRTXTMODE *******************************************************
	disables text mode
****************************************************************************/
void lcd_clrtxtmode(void) {
	lcd_mode = lcd_mode & ~LCD_CMD_SETDISPLAYMODE_TXT; // clear bit
	lcd_command(lcd_mode);
}


/***** LCD_SETGFXMODE *******************************************************
	enables graphics mode
****************************************************************************/
void lcd_setgfxmode(void) {
	lcd_mode = lcd_mode | LCD_CMD_SETDISPLAYMODE_GFX; // set bit
	lcd_command(lcd_mode);
}


/***** LCD_CLRGFXMODE *******************************************************
	disables graphics mode
****************************************************************************/
void lcd_clrgfxmode(void) {
	lcd_mode = lcd_mode & ~LCD_CMD_SETDISPLAYMODE_GFX; // clear bit
	lcd_command(lcd_mode);
}


/***** LCD_DELAY100MS *******************************************************
	100 ms delay
****************************************************************************/
void lcd_delay100ms(void) {
	asm volatile(
	"push r16"				"\n\t"
	"push r17"				"\n\t"
	"push r18"				"\n\t"

	"ldi  r16, 0x5F" 		"\n\t" // delaying 799995 cycles:
"lcd_delay100ms_1:"			"\n\t"
	"ldi  r17, 0x17"		"\n\t"
"lcd_delay100ms_2:"			"\n\t"
	"ldi  r18, 0x79"		"\n\t"
"lcd_delay100ms_3:"			"\n\t"
	"dec  r18"				"\n\t"
	"brne lcd_delay100ms_3"	"\n\t"
	"dec  r17"				"\n\t"
	"brne lcd_delay100ms_2"	"\n\t"
	"dec  r16"				"\n\t"
	"brne lcd_delay100ms_1"	"\n\t"

	"pop r18"				"\n\t"
	"pop r17"				"\n\t"
	"pop r16"
    );
}



/***** LCD_CLEARTXT *********************************************************
	clears data memory of text area
****************************************************************************/
void lcd_cleartxt(void) {
	// set address pointer to start of text area
	lcd_write(LCD_TXTHOME_LO);
	lcd_write(LCD_TXTHOME_HI);
	lcd_command(LCD_CMD_SETADDRPTR);

	// clear LCD_COLS x LCD_ROWS characters
	for(uint8_t i=0; i < LCD_COLS; i++){
		for(uint8_t k=0; k < LCD_ROWS; k++){
			lcd_txtdata(' ');
		}
	}

}


/***** LCD_CLEARGFX *********************************************************
	clears data memory of graphics area
****************************************************************************/
void lcd_cleargfx(void) {
	// set address pointer to start of graphics area
	lcd_write(LCD_GFXHOME_LO);
	lcd_write(LCD_GFXHOME_HI);
	lcd_command(LCD_CMD_SETADDRPTR);

	// clear LCD_COLS x LCD_ROWS*8 bytes
	for(uint8_t i=0; i < LCD_COLS; i++){
		for(uint8_t k=0; k < LCD_ROWS*8; k++){
			lcd_gfxdata(0x00);
		}
	}

}


/***** LCD_INIT *************************************************************
	initialises the display
****************************************************************************/
void lcd_init (void) {
	// prepare control lines (no intereference with other lines on this port)
	outp((1<<LCD_CTRL_PIN_RESET)|(1<<LCD_CTRL_PIN_CD)|(1<<LCD_CTRL_PIN_CE)|(1<<LCD_CTRL_PIN_RD)|(1<<LCD_CTRL_PIN_WR), LCD_CTRL_DDR);
	outp((1<<LCD_CTRL_PIN_RESET)|(1<<LCD_CTRL_PIN_CD)|(1<<LCD_CTRL_PIN_CE)|(1<<LCD_CTRL_PIN_RD)|(1<<LCD_CTRL_PIN_WR), LCD_CTRL_OUT);
	cbi(LCD_CTRL_OUT, LCD_CTRL_PIN_RESET);
	lcd_delay100ms();
	sbi(LCD_CTRL_OUT, LCD_CTRL_PIN_RESET);
	
	// text home address
	lcd_write(LCD_TXTHOME_LO);
	lcd_write(LCD_TXTHOME_HI);
	lcd_command(LCD_CMD_SETTXTHOME);

	// define text area
	lcd_write(LCD_COLS);
	lcd_write(0);
	lcd_command(LCD_CMD_SETTXTAREA);

	// graphics home address
	lcd_write(LCD_GFXHOME_LO);
	lcd_write(LCD_GFXHOME_HI);
	lcd_command(LCD_CMD_SETGFXHOME);

	// define graphics area
	lcd_write(LCD_COLS);
	lcd_write(0);
	lcd_command(LCD_CMD_SETGFXAREA);


	lcd_command(LCD_CMD_SETMODE); // set mode (AND/OR/XOR/text attr.)

	lcd_command(LCD_CMD_SETDISPLAYMODE); // set standard display mode (text/grpahics/cursor/blink off)

	lcd_cleartxt(); // clear text area
	lcd_cleargfx(); // clear graphics area

}


/***** LCD_PRINTBITMAP ******************************************************
	prints a bitmap
****************************************************************************/
void lcd_printbitmap (void) {

	uint8_t bitmap;
	uint8_t buffer = 0;
	uint8_t k = 0;
	uint8_t output;

	// set address pointer to start of graphics area
	lcd_write(LCD_GFXHOME_LO);
	lcd_write(LCD_GFXHOME_HI);
	lcd_command(LCD_CMD_SETADDRPTR);

	// get image data from program memory
	for(uint16_t i=0; i < 1920; i++) {
		bitmap = pgm_read_byte(&(lcd_bitmap[i]));

		// 6x8 mode: make 8->6 bit tranformation
		k++; // 1..2..3
		output = bitmap >> (k*2); // shift 2..4..6 (keep 6..4..2) bits
		output = output | buffer; // take 0..2..4 buffered bits
		lcd_gfxdata(output); // store in display RAM
		buffer = (bitmap << (6-(k*2)) )&0x3f; // prepare buffer (2..4..6 bits)

		// buffered 6 bits after 3rd tranformations -> send them to display RAM
		if (k==3) {
			lcd_gfxdata(buffer);
			buffer = 0;
			k = 0;
		}
	}

}


/***** MAIN ************************************************************************
	main function
***********************************************************************************/
int main(void) {
	
	lcd_init (); // initialise display


	// the following is just for testing purposes

	// set address pointer to start of graphics area
	lcd_write(LCD_GFXHOME_LO);
	lcd_write(LCD_GFXHOME_HI);
	lcd_command(LCD_CMD_SETADDRPTR);

//	lcd_gfxdata(0x0F); // print some bits
	lcd_setgfxmode(); // graphics mode on
	
	lcd_printbitmap(); // print a bitmap


	// set address pointer to start of text area
	lcd_write(LCD_TXTHOME_LO);
	lcd_write(LCD_TXTHOME_HI);
	lcd_command(LCD_CMD_SETADDRPTR);

	lcd_txtdata('A'); // print a character
//	lcd_settxtmode(); // graphics mode on

	return (0);
}
